"""Functions that will create transitive win-graphs from (possibly) intransitive win-graphs.

Input is a intransitive win-graph, output is a transitive win-graph."""
import random
from functools import partial

import networkx as nx


def break_random_link(vote_set):
    """While there is a cycle, breaks the cycle by removing a random edge in it."""
    win_graph = nx.DiGraph()
    win_graph.add_edges_from((vote[0], vote[1]) for vote in vote_set)

    # Keep iterating until there are no cycles remaining
    while True:
        try:
            cycle = nx.find_cycle(win_graph)
            win_graph.remove_edge(*cycle[random.randrange(0, len(cycle))])
        except nx.NetworkXNoCycle:
            break

    assert nx.is_directed_acyclic_graph(win_graph)
    return win_graph


def make_break_weakest_link(edge_to_weight):
    """Given a mapping from edges to weights, resolves intransitivity by picking cycles at random, and removing
    the weakest edge in the chosen cycle. Eventually, there are no cycles left.

    :param edge_to_weight: a dictionary mapping (winner,loser) edges to weights (floats)
    :return: a transitive vote graph
    """
    # Partial function instead of local definition so that result can be pickled
    return partial(break_weakest_link, edge_to_weight)


def break_weakest_link(edge_to_weight, vote_set):
    """While there is a cycle, breaks the cycle by removing the weakest edge in it."""
    win_graph = nx.DiGraph()
    win_graph.add_edges_from((vote[0], vote[1]) for vote in vote_set)

    def weakest(edges):
        return min(edges, key=lambda e: edge_to_weight[e])

    # Keep iterating until there are no cycles remaining
    while True:
        try:
            cycle = nx.find_cycle(win_graph)
            win_graph.remove_edge(*weakest(cycle))
        except nx.NetworkXNoCycle:
            break

    assert nx.is_directed_acyclic_graph(win_graph)
    return win_graph


def make_add_edges_in_order(edge_to_weights):
    # Partial function instead of local definition so that result can be pickled
    return partial(add_edges_in_order, edge_to_weights)


def add_edges_in_order(edge_to_weights, vote_set):
    """Adds edges in order of weight, never adding edges that would create a cycle."""
    win_graph = nx.DiGraph()
    ordered_votes = sorted(vote_set, key=lambda e: edge_to_weights[(e[0], e[1])], reverse=True)
    for c1, c2, result in ordered_votes:
        try:
            win_graph.add_edge(c1, c2)
            cycles = nx.find_cycle(win_graph)
            # If we hit this line, a NoCycle exception was not thrown, therefore there is a cycle and we have to
            # remove an edge from it
            win_graph.remove_edge(c1, c2)
        except nx.NetworkXNoCycle:
            continue

    assert nx.is_directed_acyclic_graph(win_graph)
    return win_graph



import math

import networkx as nx
import pygraphviz as pgv


def write_beatgraph(
    victory_graph: nx.DiGraph, rankings: list, win_ratio: dict, output_filename: str = ""
):
    """Draws the beatgraph of the given matchups to a file

    :param victory_graph: a nx_graph, with nodes as candidates and edges (u, v)
    representing a victory where u beat v. Each edge must have the edge_data attribute
    "margin", explaining by how much u beat v.
    :param win_ratio: a dictionary mapping node names to win ratios in [0, 1]
    :param rankings: an array, with the same elements as `nx_graph.nodes`, with the ranking of the nodes.
    :param output_filename: the name of the file to write to
    """
    win_edges = sorted(
        [edge for edge in victory_graph.edges],
        key=lambda x: victory_graph.get_edge_data(x[0], x[1])["margin"],
        reverse=True,
    )

    new_graph = nx.DiGraph()
    new_graph.add_nodes_from(victory_graph.nodes)
    for (u, v) in win_edges:
        if v not in nx.descendants(new_graph, u):
            new_graph.add_edge(u, v)

    to_remove = []
    e = [x for x in new_graph.edges]
    for (u, v) in e:
        new_graph.remove_edge(u, v)
        has_other_path = nx.has_path(new_graph, u, v)

        if not has_other_path:
            new_graph.add_edge(u, v)

    new_graph.remove_edges_from(to_remove)

    G = pgv.AGraph(directed=True)
    G.node_attr["style"] = "filled"

    base_color = "0.520 0.500 "
    for n in victory_graph.nodes:
        color = 1 - rankings.index(n) / len(rankings) / 2
        G.add_node(
            n,
            fillcolor=base_color + ("%.3f" % color),
            height=math.sqrt(win_ratio[n]) * 2,
            width=math.sqrt(win_ratio[n]) * 2,
        )

    for u, v in new_graph.edges:
        edge_data = victory_graph.get_edge_data(u, v)
        G.add_edge(
            u,
            v,
            penwidth=math.sqrt(edge_data["margin"] / 4),
            xlabel="%.2f" % (edge_data["margin"]),
            labelfloat="false",
        )

    G.layout(prog="dot")
    if output_filename:
        G.draw(output_filename, format="png")
    else:
        return G.draw(format="png", prog="dot")


if __name__ == "__main__":
    g = nx.DiGraph()
    g.add_node("python")
    g.add_node("java")
    g.add_node("scala")
    g.add_node("fortran")
    g.add_node("cobol")

    g.add_edge("scala", "java", margin=0.60)
    g.add_edge("scala", "python", margin=0.10)
    g.add_edge("scala", "cobol", margin=0.12)
    g.add_edge("scala", "fortran", margin=0.100)

    g.add_edge("python", "java", margin=0.30)
    g.add_edge("python", "cobol", margin=0.20)
    g.add_edge("python", "fortran", margin=0.80)

    g.add_edge("java", "cobol", margin=0.80)
    g.add_edge("java", "c", margin=0.23)

    g.add_edge("c", "cobol", margin=0.80)

    g.add_edge("fortran", "cobol", margin=0.10)

    rankings = ["scala", "python", "java", "c", "fortran", "cobol"]

    win_ratios = {"scala": 1, "python": 0.8, "java": 0.6, "c": 0.4, "fortran": 0.2, "cobol": 0.1}

    output_filename = "viz.png"

    write_beatgraph(g, rankings, win_ratios, output_filename)



"""Functions that will add missing items into a partial ranking. All functions have the following signature:

Input is a (possibly) partial, but transitive win-graph, and a set of nodes to insert into that win-graph. The output is a
complete, fully connected, transitive win-graph."""
import random
from functools import partial

import networkx as nx

from socialchoice import util
from socialchoice import PairwiseBallotBox

# TODO this file has duplicate code in each of the functions, converting candidates to `to_add`
#      and also roundtripping to rankings
from socialchoice.util import candidates_in_ranked_choice_ballot


def place_randomly(win_graph: nx.DiGraph, candidates: set) -> nx.DiGraph:
    """Inserts each candidate to a random place in the ranking."""
    ranking = _graph_to_ranking(win_graph)
    to_add = candidates.difference(set(win_graph.nodes))
    for item in to_add:
        ranking.insert(random.randrange(0, len(ranking) + 1), item)
    return _ranking_to_graph(ranking)


def add_all_at_beginning(win_graph: nx.DiGraph, candidates: set) -> nx.DiGraph:
    """Adds all candidates as winning against everyone."""
    to_add = candidates.difference(set(win_graph.nodes))
    ranking = _graph_to_ranking(win_graph)
    ranking.insert(0, to_add)
    return _ranking_to_graph(ranking)


def add_all_at_end(win_graph: nx.DiGraph, candidates: set) -> nx.DiGraph:
    """Adds all candidates to the end of the ranking."""
    to_add = candidates.difference(set(win_graph.nodes))
    ranking = _graph_to_ranking(win_graph)
    ranking.append(to_add)
    return _ranking_to_graph(ranking)


def add_random_edges(win_graph: nx.DiGraph, candidates: set) -> nx.DiGraph:
    """Chooses a random pair of nodes that aren’t connected to each other, and then connects them,
    never adding edges that would result in a cycle, until the graph is a complete win-graph.
    """
    to_add = candidates.difference(set(win_graph.nodes))
    win_graph = win_graph.copy()
    candidates = set(win_graph.nodes).union(to_add)
    edge_list = []
    for u in candidates:
        for v in candidates:
            if u != v:
                edge_list.append((u, v))

    random.shuffle(edge_list)

    # Existing edges don't have to be checked
    for edge in win_graph.edges:
        edge_list.remove(edge)

    for c1, c2 in edge_list:
        try:
            win_graph.add_edge(c1, c2)
            nx.find_cycle(win_graph)
            win_graph.remove_edge(c1, c2)
        except nx.NetworkXNoCycle:
            pass

    return win_graph


def make_add_edges_by_win_ratio(edges_to_win_ratio):
    """Given a list of edges by win ratio, creates a function that will resolve incompleteness by
    adding non-cycle-creating edges from the list until the graph is complete.

    :param edges_to_win_ratio: the list of edges as 2-tuples of (winner, loser),
                               
    ordered by win rate, highest first
    :return:
    """
    edges_by_win_ratio = sorted(
        edges_to_win_ratio, key=lambda e: edges_to_win_ratio[e], reverse=True
    )
    # Partial function instead of local definition so that result can be pickled
    return partial(add_edges_by_win_ratio, edges_by_win_ratio)


def add_edges_by_win_ratio(
    edges_by_win_ratio, win_graph: nx.DiGraph, candidates: set
) -> nx.DiGraph:
    """Adds edges into the win-graph in order of the win ratios of those matchups in the entire voting set,
    only adding edges that will not create cycles."""
    to_add = candidates.difference(set(win_graph.nodes))
    for (c1, c2) in edges_by_win_ratio:
        try:
            win_graph.add_edge(c1, c2)
            nx.find_cycle(win_graph)
            win_graph.remove_edge(c1, c2)
        except nx.NetworkXNoCycle:
            pass

    assert all(candidate in win_graph.nodes for candidate in to_add)
    return win_graph


def _graph_to_ranking(g: nx.DiGraph) -> list:
    return list(nx.topological_sort(g))


def _ranking_to_graph(r: list) -> nx.DiGraph:
    ballots = util.ranking_to_pairwise_ballots(r)
    candidates = candidates_in_ranked_choice_ballot(r)
    return PairwiseBallotBox(ballots, candidates).get_victory_graph()



import scipy.stats


def kendalls_tau(ranking1, ranking2):
    def ranks_of_candidates(ranking):
        r = {}
        ri = 0

        for tie_set in ranking:
            if isinstance(tie_set, set):
                ri += len(tie_set) / 2
                for candidate in tie_set:
                    r[candidate] = ri
                ri += len(tie_set) / 2
            else:
                candidate = tie_set
                ri += 1
                r[candidate] = ri
        return r

    ranks1 = ranks_of_candidates(ranking1)
    ranks2 = ranks_of_candidates(ranking2)

    x = [ranks1[item] for item in ranking1]
    y = [ranks2[item] for item in ranking1]
    return scipy.stats.kendalltau(x, y)


def num_inversions(ranking1, ranking2):
    """This is a measure of how far apart two rankings are, by the number of inversions. An inversion is a
    swapping of two elements, so [1,2,3] and [1,3,2] have one inversion, while [1,2,3] and [3,1,2] have two.

    :raises ValueError: if called with rankings that don't contain the same elements.
    """
    ranking1 = list(ranking1)
    ranking2 = list(ranking2)

    if set(ranking1) != set(ranking2):
        raise ValueError(f"Rankings must contain same elements, {ranking1} and {ranking2} differ")

    inversions = 0
    for index, item1 in enumerate(ranking1):
        for item2 in ranking1[index + 1 :]:
            if ranking2.index(item1) >= ranking2.index(item2):
                inversions += 1

    return inversions



import random

import networkx as nx
import pytest
from hypothesis import given, note
from hypothesis import strategies as st

from socialchoice.ballot import PairwiseBallotBox
from socialchoice.election import Election
from socialchoice.ranking_similarity import num_inversions
from socialchoice.util import ranking_with_all_sets, ranking_to_pairwise_ballots


def test_ranking_with_all_sets():
    assert ranking_with_all_sets([1, 2, 3]) == [{1}, {2}, {3}]
    assert ranking_with_all_sets([1, {2, 3}]) == [{1}, {2, 3}]
    assert ranking_with_all_sets([{1}, {2, 3}]) == [{1}, {2, 3}]


def test_ranking_to_pairwise_ballots():
    assert set(ranking_to_pairwise_ballots([1, 2])) == {(1, 2, "win")}
    assert set(ranking_to_pairwise_ballots([1, 2, 3])) == {
        (1, 2, "win"),
        (1, 3, "win"),
        (2, 3, "win"),
    }
    assert set(ranking_to_pairwise_ballots([1, {2, 3}])) == {
        (1, 2, "win"),
        (1, 3, "win"),
        (2, 3, "tie"),
    }
    assert set(ranking_to_pairwise_ballots([{1, 2}, {3, 4}])) == {
        (1, 2, "tie"),
        (1, 3, "win"),
        (1, 4, "win"),
        (2, 3, "win"),
        (2, 4, "win"),
        (3, 4, "tie"),
    }


@given(st.lists(st.integers(), unique=True, min_size=2))
def test_pairwise_ballots_roundtrip(ranking):
    """Converting a ranking to pairwise ballots and then toposorting will produce the same ranking"""

    pairwise = ranking_to_pairwise_ballots(ranking)
    pairwise_ranked_pairs = list(
        nx.topological_sort(PairwiseBallotBox(pairwise).get_victory_graph())
    )
    assert pairwise_ranked_pairs == ranking



from itertools import combinations


def ranking_with_all_sets(r):
    out = []
    for item in r:
        if isinstance(item, set):
            out.append(item)
        else:
            out.append({item})
    return out


def ranking_to_pairwise_ballots(ranking):
    ranking = ranking_with_all_sets(ranking)
    # to convert to pairwise ballots, we'll have to look at each ranking
    pairwise_ballots = []
    # and within each set in that ranking, all pairs of elements (combinations of size 2) are tied
    for candidate_set in ranking:
        for candidate1, candidate2 in combinations(candidate_set, 2):
            pairwise_ballots.append((candidate1, candidate2, "tie"))

    # For each candidate set, and every candidate set after them in the ranking, the first set wins against
    # each set in the second
    for i, winner_candidate_set in enumerate(ranking):
        for losing_candidate_set in ranking[i + 1 :]:
            # (but because they are both sets, we have to iterate over each candidate in each)
            for winning_candidate in winner_candidate_set:
                for losing_candidate in losing_candidate_set:
                    pairwise_ballots.append((winning_candidate, losing_candidate, "win"))

    return pairwise_ballots


def candidates_in_ranked_choice_ballot(ballot):
    candidate_set = set()
    for item in ballot:
        if isinstance(item, set):
            # If there is a set, it is a tie, so process it one element at a time.
            for elem in item:
                if elem in candidate_set:
                    raise ValueError(f"Candidate {elem} appears multiple times in {ballot}")
                candidate_set.add(elem)
        else:
            # If it's not a set, we can just process one at a time.
            if item in candidate_set:
                raise ValueError(f"Candidate {item} appears multiple times in {ballot}")
            candidate_set.add(item)
    return candidate_set



from socialchoice import Election, PairwiseBallotBox

empty_election = Election(PairwiseBallotBox([]))
example_votes = PairwiseBallotBox(
    [[0, 1, "win"], [3, 2, "loss"], [2, 3, "win"], [0, 3, "tie"], [3, 0, "win"]]
)


def test_get_ranked_pairs_ranking():
    """Tests that ranked_pairs on a pairwise ballot box produces the correct outcome."""
    assert empty_election.ranking_by_ranked_pairs() == []

    e_1 = Election(example_votes)
    assert e_1.ranking_by_ranked_pairs() == [2, 3, 0, 1]


def test_get_win_ratio():
    assert empty_election.ranking_by_win_ratio() == []

    e_1 = Election(example_votes)
    assert e_1.ranking_by_win_ratio(include_score=True) == [
        (2, 1.0),
        (0, 0.5),
        (3, 1 / 3),
        (1, 0.0),
    ]


def test_get_win_tie_ratio():
    assert empty_election.ranking_by_win_tie_ratio() == []

    e_1 = Election(example_votes)
    assert e_1.ranking_by_win_tie_ratio(include_score=True) == [
        (2, 1.0),
        (0, 2 / 3),
        (3, 0.5),
        (1, 0.0),
    ]


def test_flatten_ties():
    election = Election(PairwiseBallotBox([(1, 2, "win"), (3, 4, "win")], candidates={1, 2, 3, 4}))
    ranking = election.ranking_by_win_ratio(group_ties=True, include_score=True)
    assert ranking == [[(1, 1.0), (3, 1.0)], [(2, 0.0)]]



from setuptools import setup

setup(
    name="socialchoice",
    version="0.0.0",
    packages=["socialchoice", "socialchoice.pairwisecollapse"],
    package_dir={"": "src"},
    url="https://github.com/julian-zucker/socialchoice",
    license="Apache 2.0",
    author="Julian Zucker",
    author_email="julian.zucker@gmail.com",
    description="Social Choice Theory in Python",
    install_requires=["networkx", "pygraphviz", "pytest", "scipy"],
)



"""
Collapse pairwise votes into ranked choice votes.
"""


def pairwise_collapse(
    pairwise_votes, candidates, intransitivity_resolver, incompleteness_resolver
) -> list:
    """
    Converts a set of pairwise votes into a ranking over all of the candidates.

    :param pairwise_votes: The pairwise votes, as a list of 2-tuples
    :param candidates: The list of candidates in the elections
    :param intransitivity_resolver: function used to resolve intransitivity
    :param incompleteness_resolver: function used to complete the transitive sub-graph
    """
    transitive_votes = intransitivity_resolver(pairwise_votes)
    return incompleteness_resolver(transitive_votes, candidates)



"""
This file holds the complete interface for this package, as everything in here can be imported as
socialchoice.{whatever} instead of socialchoice.{filename}.{whatever}

This allows us to factor out the package into multiple files while keeping a consistent user-facing API.
"""

from socialchoice.ballot import *

from socialchoice.election import Election

import socialchoice.pairwisecollapse.pairwise_collapse as pairwise_collapse

import socialchoice.util



import random

import pytest
from hypothesis import given, note, strategies as st

from socialchoice.ranking_similarity import num_inversions


def test_num_inversions():
    assert num_inversions([], []) == 0
    assert num_inversions([1], [1]) == 0
    assert num_inversions([1, 2], [1, 2]) == 0
    assert num_inversions([0, 1], [1, 0]) == 1
    assert num_inversions([1, 2], [2, 1]) == 1
    assert num_inversions([2, 1], [1, 2]) == 1
    assert num_inversions([1, 2, 3], [3, 2, 1]) == 3
    assert num_inversions([3, 2, 1], [1, 2, 3]) == 3


def test_num_inversions_assertion_error_if_different_sets():
    with pytest.raises(ValueError):
        num_inversions([1], [2])

    with pytest.raises(ValueError):
        num_inversions([3], [3, 4])

    with pytest.raises(ValueError):
        num_inversions([1, 2], [2])


@pytest.mark.hypothesis
@pytest.mark.slow
@given(st.lists(st.integers(), unique=True))
def test_inversions_between_list_and_inverse_triangle(lst):
    """The number of inversions between a list and it's inverse should be the `len(l)`th triangle number."""
    note(f"List: {lst}, reversed: {list(reversed(lst))}")
    assert num_inversions(lst, reversed(lst)) == (len(lst) * (len(lst) - 1)) // 2


@pytest.mark.hypothesis
@pytest.mark.slow
@given(st.lists(st.integers(), unique=True))
def test_no_inversions_identical_list(lst):
    note(f"List: {lst}")
    assert num_inversions(lst, lst) == 0


@pytest.mark.hypothesis
@pytest.mark.slow
@given(st.lists(st.integers(), unique=True))
def test_inversions_order_irrelevant(lst):
    lst1 = lst
    lst2 = random.sample(lst, len(lst))

    note(f"Lists: {lst1, lst2}")
    assert num_inversions(lst1, lst2) == num_inversions(lst2, lst1)



import functools

import networkx as nx

from socialchoice.ballot import BallotBox


def optional_score(ranking_method):
    """Adds a flag to include or remove the score from a ranking method that returns a list of 2-tuples,
    where the first element is the candidate, and the second element is a score.

    :param ranking_method: a method on an Election
    :return: the same ranking method, but with a flag `include_score`, which defaults to False. If set to true, the
    function returns a list of 2-tuples instead of just a ranking
    """
    # This unfortunately has to be defined above Election to be used in Election

    def nest_ties(ranking):
        """
        >>> nest_ties([("a", 1), ("b", 1), ("c", 0)])
        [[("a", 1), ("b", 1)], [("c", 0)]]

        :param ranking: A list of 2-tuples of (item, score)
        :return: a list of lists of 2-tuples, where each list in the outer list contains only items with the same score.
        """
        if len(ranking) == 0:
            return ranking

        out = []
        current_tie = [ranking[0]]
        for item, score in ranking[1:]:
            # There are no items in the current tied set, or the score of the current item is the same
            if current_tie[0][0] == score:
                current_tie.append((item, score))
            else:
                out.append(current_tie)
                current_tie = [(item, score)]
        return out

    @functools.wraps(ranking_method)
    def wrapped_ranking_method(self, include_score=False, group_ties=False):
        ranking = ranking_method(self)

        if include_score and group_ties:
            return nest_ties(ranking)
        elif include_score and not group_ties:
            return ranking
        elif not include_score and group_ties:
            return [item for tie_group in nest_ties(ranking) for item, score in tie_group]
        elif not include_score and not group_ties:
            return [item for item, score in ranking]

    return wrapped_ranking_method


class Election:
    """Given a ballot box, allows you to run social choice methods on the ballot box."""

    def __init__(self, ballot_box: BallotBox):
        self.ballot_box = ballot_box

    def ranking_by_ranked_pairs(self) -> list:
        matchups = self.ballot_box.get_victory_graph()

        g = nx.DiGraph()
        g.add_nodes_from(matchups.nodes)

        edges = [(u, v, matchups.get_edge_data(u, v)) for (u, v) in matchups.edges]
        edges.sort(key=lambda x: x[2]["margin"], reverse=True)
        for u, v, data in edges:
            g.add_edge(u, v, **data)
            try:
                nx.find_cycle(g)
                g.remove_edge(u, v)
            except nx.NetworkXNoCycle:
                pass

        assert nx.is_directed_acyclic_graph(g)
        return list(nx.topological_sort(g))

    @optional_score
    def ranking_by_copeland(self) -> list:
        g = self.ballot_box.get_victory_graph()
        result = sorted(
            [(n, g.out_degree(n) - g.in_degree(n)) for n in g.nodes],
            key=lambda x: x[1],
            reverse=True,
        )
        return result

    @optional_score
    def ranking_by_minimax(self) -> list:
        g = self.ballot_box.get_matchup_graph()
        return sorted(
            g.nodes, key=lambda n: max(g.get_edge_data(u, v)["margin"] for u, v in g.in_edges(n))
        )

    @optional_score
    def ranking_by_win_ratio(self) -> list:
        matchups = self.ballot_box.get_matchups()

        wins_and_ties_vs_losses = {}
        for candidate in matchups:
            wins_and_ties_vs_losses[candidate] = [0, 0]
            for matchup in matchups[candidate]:
                wins = matchups[candidate][matchup]["wins"]
                losses = matchups[candidate][matchup]["losses"]
                wins_and_ties_vs_losses[candidate][0] += wins
                wins_and_ties_vs_losses[candidate][1] += losses

        ratios = [
            (candidate, (x[0] / ((x[0] + x[1]) or float("inf"))))
            for candidate, x in wins_and_ties_vs_losses.items()
        ]
        return sorted(ratios, key=lambda x: x[1], reverse=True)

    @optional_score
    def ranking_by_win_tie_ratio(self) -> list:
        matchups = self.ballot_box.get_matchups()

        wins_and_ties_vs_losses = {}
        for candidate in matchups:
            wins_and_ties_vs_losses[candidate] = [0, 0]
            for matchup in matchups[candidate]:
                wins = matchups[candidate][matchup]["wins"]
                ties = matchups[candidate][matchup]["ties"]
                losses = matchups[candidate][matchup]["losses"]
                wins_and_ties_vs_losses[candidate][0] += wins + ties
                wins_and_ties_vs_losses[candidate][1] += losses

        ratios = [
            (candidate, (x[0] / ((x[0] + x[1]) or float("inf"))))
            for candidate, x in wins_and_ties_vs_losses.items()
        ]
        return sorted(ratios, key=lambda x: x[1], reverse=True)



import networkx as nx
from more_itertools import flatten

import socialchoice.util as util


class BallotBox:
    """An interface for the features of ballot boxes"""

    def get_victory_graph(self) -> nx.DiGraph:
        """
        A victory graph is a matchup graph of each candidate, but with only edges for wins.
        Every out-edge represents a win over another candidate. Two candidates will not have
        any edge between them if there is a perfect tie (by win ratio).

        See `get_matchup_graph` for a description of the attributes on nodes and edges.
         """
        pass

    def get_matchup_graph(self) -> nx.DiGraph:
        """
        A matchup graph is a fully-connected graph, with each out-edge corresponding to a matchup
        and each in-edge corresponding to the same matchup, but with wins and losses flipped.
        An edge `(a, b)` from a to b will have `wins` corresponding to the number of wins `a`
        has over `b`, while the edge `(b, a)` will have `wins` corresponding to the number of wins
        `b` has over `a`.

        Every edge has the attributes `wins` (described above), `losses`, `ties`, and `margin`.
        `losses` and `ties` are self explanatory, simply the number of losses or ties between
        the two candidates. `margin` is the ratio of wins to overall votes.
        """
        pass

    def get_matchups(self) -> dict:
        """This matchup shows the number of wins and losses each candidate has against each other.
        The shape of the final dictionary returned is:

         >>> {\
            "candidate1": { \
                 {"candidate2": {"win": 12, "loss": 10, "tie": 5}}, \
                 {"candidate3": {"win": 3,  "loss": 23, "tie": 9}}, \
                 # potentially many more \
             }, \
             "candidate2": { \
                 {"candidate1": {"win": 10, "loss": 12, "tie": 5}}, \
                 {"candidate3": {"win": 23, "loss":  3, "tie": 9}}, \
                 # potentially many more \
             },\
             # and so on\
          }

        :return: a matchup mapping, as described above
        """


class PairwiseBallotBox(BallotBox):
    """Stores ballots in pairwise form, as in: ["Alice",  "Bob", "win"]"""

    def __init__(self, votes, candidates=None):
        """
        Creates a new PairwiseBallotBox. This shows that

        :param votes: An array of votes, where a vote is any indexable object of length 3
        :param candidates: None, meaning to infer the candidate set from the votes, or a
                           collection of the candidates that were being voted on in this election.
        :raises InvalidVoteShapeException: if given any vote with length != 3
        """
        votes = list(votes)
        self.ballots = self.__ensure_valid_votes(votes)
        self.candidates = candidates or self.__get_all_candidates_from_votes(self.ballots)

    @staticmethod
    def __get_all_candidates_from_votes(votes) -> set:
        """:return: All the candidates mentioned in the votes"""
        return {vote[0] for vote in votes} | {vote[1] for vote in votes}

    @staticmethod
    def __ensure_valid_votes(votes: iter) -> list:
        for vote in votes:
            if not len(vote) == 3:
                raise InvalidBallotDataException(
                    "Expected a vote of length three, got " + str(vote)
                )

            if not vote[2] in {"win", "loss", "tie"}:
                raise InvalidBallotDataException(
                    """Expected type to be one of {"win", "loss", "tie"}, got""" + str(vote)
                )

        return list(votes)

    def get_victory_graph(self) -> nx.DiGraph:
        matchups = self.get_matchup_graph()
        edges_to_remove = []
        for u, v in matchups.edges:
            margin_u_to_v = matchups.get_edge_data(u, v)["margin"]
            margin_v_to_u = matchups.get_edge_data(v, u)["margin"]

            # u->v has lower win margin than v->u, keep the "winning edge" with higher win ratio
            # this will remove both if there is a perfect tie - but that's okay
            if margin_u_to_v <= margin_v_to_u:
                edges_to_remove.append((u, v))

        matchups.remove_edges_from(edges_to_remove)
        return matchups

    def get_matchup_graph(self) -> nx.DiGraph:
        matchups = self.get_matchups()
        ids = matchups.keys()

        g = nx.DiGraph()
        g.add_nodes_from(ids)
        for candidate1 in matchups:
            for candidate2 in matchups[candidate1]:
                candidate1to2 = matchups[candidate1][candidate2]
                total_votes_candidate1to2 = sum(candidate1to2.values())
                if total_votes_candidate1to2 != 0:
                    g.add_edge(
                        candidate1,
                        candidate2,
                        wins=candidate1to2["wins"],
                        losses=candidate1to2["losses"],
                        ties=candidate1to2["ties"],
                        margin=candidate1to2["wins"] / total_votes_candidate1to2,
                    )

                candidate2to1 = matchups[candidate2][candidate1]
                total_votes_candidate2to1 = sum(candidate2to1.values())
                if total_votes_candidate2to1 != 0:
                    g.add_edge(
                        candidate2,
                        candidate1,
                        wins=candidate2to1["wins"],
                        losses=candidate2to1["losses"],
                        ties=candidate2to1["ties"],
                        margin=candidate2to1["wins"] / total_votes_candidate2to1,
                    )

        return g

    def get_matchups(self) -> dict:
        matchups = {}
        for candidate in self.candidates:
            matchups[candidate] = {}

        for candidate1 in self.candidates:
            for candidate2 in self.candidates:
                if candidate1 == candidate2:
                    continue
                matchups[candidate1][candidate2] = {"wins": 0, "losses": 0, "ties": 0}
                matchups[candidate2][candidate1] = {"wins": 0, "losses": 0, "ties": 0}

        for vote in self.ballots:
            candidate1, candidate2, result = vote
            if result == "win":
                matchups[candidate1][candidate2]["wins"] += 1
                matchups[candidate2][candidate1]["losses"] += 1
            if result == "loss":
                matchups[candidate1][candidate2]["losses"] += 1
                matchups[candidate2][candidate1]["wins"] += 1
            if result == "tie":
                matchups[candidate1][candidate2]["ties"] += 1
                matchups[candidate2][candidate1]["ties"] += 1

        return matchups


class RankedChoiceBallotBox(BallotBox):
    def __init__(self, ballots, candidates=None):
        """Creates a RankedChoiceBallotBox from the given ballots. Each ballot must be a list, where
        each element is either a candidate or a set of candidates. A single candidate in a ballot
        represents that candidate being at that position, and a set represents a tie for that
        position.

        Each ballot must contain every candidate found in candidates, or if candidates is not
        provided, the candidates mentioned in every other ballot.

        :param ballots: a list of ballots, as described above.
        :param candidates: the set of candidates being voted on. Inferred from ballots if not
        provided.
        """
        self.ballots = self.__ensure_valid_ballots(ballots, candidates)
        self.ballots_all_sets = self.__convert_to_sets(self.ballots)

        # We want to convert to pairwise ballots because there's no use reimplementing the code in
        # PairwiseBallotBox for rankings, we can just convert a ranking to its constituent
        # pairwise preferences and create our own PairwiseBallotBox that we can forward
        # requests for pairwise-result based rankings to.
        pairwise_ballots = flatten(
            util.ranking_to_pairwise_ballots(ballot) for ballot in self.ballots_all_sets
        )

        self.pairwise_ballot_box = PairwiseBallotBox(pairwise_ballots)

    def get_victory_graph(self) -> nx.DiGraph:
        return self.pairwise_ballot_box.get_victory_graph()

    def get_matchup_graph(self) -> nx.DiGraph:
        return self.pairwise_ballot_box.get_matchup_graph()

    def get_matchups(self) -> dict:
        return self.pairwise_ballot_box.get_matchups()

    def __ensure_valid_ballots(self, ballots, candidates):
        if not len(ballots):
            raise InvalidBallotDataException(
                "Cannot create RankedChoiceBallotBox with empty ballot list"
            )
        for ballot in ballots:
            if not isinstance(ballot, list):
                raise InvalidBallotDataException(
                    f"Ballots must be a collection of lists, one ballot was {ballot}"
                )

        # Need to either get or infer the candidate set to validate that ballots are full, and to
        # ensure that each ballot doesn't contain elements not in the candidate set.
        if candidates:
            candidate_set = set(candidates)
            if not len(candidates) == len(candidate_set):
                raise InvalidElectionDataException(f"Duplicate candidates in {candidates}")
        else:
            candidate_set = set()
            for ballot in ballots:
                try:
                    candidate_set |= util.candidates_in_ranked_choice_ballot(ballot)
                except ValueError as e:
                    raise InvalidBallotDataException(e)

        for ballot in ballots:
            try:
                ballot_contents = util.candidates_in_ranked_choice_ballot(ballot)
            except ValueError as e:
                raise InvalidBallotDataException(e)

            if ballot_contents != candidate_set:
                raise InvalidBallotDataException(
                    f"Ballot {ballot} did not contain exactly {candidate_set}."
                )

        return ballots

    def __convert_to_sets(self, ballots):
        """Takes ballots, which may include single items at rankings which do not have ties, into a list of sets."""
        return [util.ranking_with_all_sets(b) for b in ballots]


class InvalidElectionDataException(Exception):
    """Raised if there is invalid data somewhere other than the ballots."""


class InvalidBallotDataException(Exception):
    """Raised if a ballot has invalid data, for example, contains a candidate not in the list of candidates."""
